{ File   : BIFF.pas
  Author : Deniz Oezmen
  Created: 2007-03-18, based on TM.pas created 2007-02-14
  Changed: 2009-04-12

  The Longest Journey BIFF converter.
}
unit BIFF;

interface

uses
  Classes;

type
  TBIFFCallback = procedure(const Frac: Double);

function ConvertBIFF(const BIFFName: string; var ResourceText: TStringList;
  const BIFFCallback: TBIFFCallback): Boolean;

implementation

uses
  TLJ_FS,
  SysUtils;

const
  BIFFHeader_ID      = 'BIFF';
  Block_BeginMarker  = $f0f0f0f0;
  Block_EndMarker    = $0f0f0f0f;
  BlockType_Image    = $02faf080;
  BlockType_Palette  = $02faf082;
  BlockType_Lister   = $05a4aa89;
  BlockType_Object   = $05a4aa8d;
  BlockType_Material = $05a4aa8e;

type
  TBIFFHeader    = packed record
                     ID       : array[0..3] of Char;
                     Version,   // 1 or 2 allowed
                     Unknown1,
                     Unknown2,
                     NumBlocks: Longword;
                   end;

  TBlockHeader   = packed record
                     BeginMarker,
                     TypeID,
                     Unknown1   : Longword;
                     DataSize   : Longint;
                     Unknown2   : Longword; //only exists in version 2
                   end;

  TBlockTrailer  = packed record
                     EndMarker,
                     NumSubBlocks: Longword;
                   end;

function OpenBIFF(var BIFFFile: TFS_Handle; BIFFName: string): Boolean;
begin
  Result := FS_OpenFile(BIFFName, BIFFFile)
end;

function ReadHeader(var BIFFFile: TFS_Handle;
  out BIFFHeader: TBIFFHeader): Boolean;
var
  BytesRead: Longword;
begin
  BytesRead := FS_BlockRead(BIFFFile, BIFFHeader, SizeOf(TBIFFHeader));
  ReadHeader := (SizeOf(TBIFFHeader) = BytesRead) and
    (BIFFHeader.ID = BIFFHeader_ID) and (BIFFHeader.Version in [1, 2])
end;

function ReadString(var BIFFFile: TFS_Handle): string;
var
  Size: Word;
begin
  FS_BlockRead(BIFFFile, Size, SizeOf(Size));
  SetLength(Result, Size);
  FS_BlockRead(BIFFFile, Result[1], Size);
  Result := string(PChar(Result))
end;

function ReadSingle(var BIFFFile: TFS_Handle): Single;
begin
  FS_BlockRead(BIFFFile, Result, SizeOf(Result))
end;

function ReadLongint(var BIFFFile: TFS_Handle): Longint;
begin
  FS_BlockRead(BIFFFile, Result, SizeOf(Result))
end;

procedure ProcessImage(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
begin
  ResourceText.Add(Format('%-*sTexture {', [IndLevel, '']));
  ResourceText.Add(Format('%-*s  [image data]', [IndLevel, '']));
  FS_Seek(BIFFFile, FS_FilePos(BIFFFile) + BlockHeader.DataSize)
end;

procedure ProcessPalette(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
begin
  ResourceText.Add(Format('%-*sPalette {', [IndLevel, '']));
  ResourceText.Add(Format('%-*s  [palette data]', [IndLevel, '']));
  FS_Seek(BIFFFile, FS_FilePos(BIFFFile) + BlockHeader.DataSize)
end;

procedure ProcessLister(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
begin
  ResourceText.Add(Format('%-*sLister {', [IndLevel, '']));
  ResourceText.Add(Format('%-*s  ListerName = "%s"',
    [IndLevel, '', ReadString(BIFFFile)]))
end;

procedure ProcessObject(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
var
  OldSep: Char;
  Size  : Longint;
begin
  OldSep := DecimalSeparator;
  DecimalSeparator := '.';

  Size := FS_FilePos(BIFFFile);

  ResourceText.Add(Format('%-*sObject {', [IndLevel, '']));

  Size := BlockHeader.DataSize - (FS_FilePos(BIFFFile) - Size);

  ResourceText.Add(Format('%-*s  [%d bytes of unknown data]',
    [IndLevel, '', Size]));
  FS_Seek(BIFFFile, FS_FilePos(BIFFFile) + Size);

  DecimalSeparator := OldSep
end;

procedure ProcessMaterial(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
var
  OldSep: Char;
  Size  : Longint;
begin
  OldSep := DecimalSeparator;
  DecimalSeparator := '.';

  Size := FS_FilePos(BIFFFile);

  ResourceText.Add(Format('%-*sMaterial {', [IndLevel, '']));
  ResourceText.Add(Format('%-*s  MaterialName = "%s"',
    [IndLevel, '', ReadString(BIFFFile)]));
  ResourceText.Add(Format('%-*s  TextureName = "%s"',
    [IndLevel, '', ReadString(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown1 = "%s"',
    [IndLevel, '', ReadString(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown2 = "%s"',
    [IndLevel, '', ReadString(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown3 = %d',
    [IndLevel, '', ReadLongint(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown4 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown5 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown6 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown7 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown8 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown9 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown10 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown11 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown12 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown13 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));
  ResourceText.Add(Format('%-*s  Unknown14 = %1.7f',
    [IndLevel, '', ReadSingle(BIFFFile)]));

  Size := BlockHeader.DataSize - (FS_FilePos(BIFFFile) - Size);

  ResourceText.Add(Format('%-*s  [further %d bytes of unknown data]',
    [IndLevel, '', Size]));
  FS_Seek(BIFFFile, FS_FilePos(BIFFFile) + Size);

  DecimalSeparator := OldSep
end;

procedure ProcessUnknown(var BIFFFile: TFS_Handle;
  const BlockHeader: TBlockHeader; IndLevel: Longword;
  var ResourceText: TStringList);
begin
  ResourceText.Add(Format('%-*sUnknown chunk type %.8x {',
    [IndLevel, '', BlockHeader.TypeID]));
  ResourceText.Add(Format('%-*s[chunk data, size: %d bytes]',
    [IndLevel + 2, '', BlockHeader.DataSize]));
  FS_Seek(BIFFFile, FS_FilePos(BIFFFile) + BlockHeader.DataSize)
end;

procedure ProcessBlock(var BIFFFile: TFS_Handle; const BIFFHeader: TBIFFHeader;
  IndLevel: Longword; var ResourceText: TStringList);
var
  i           : Longint;
  BlockHeader : TBlockHeader;
  BlockTrailer: TBlockTrailer;
begin
  if BIFFHeader.Version = 2 then
    FS_BlockRead(BIFFFile, BlockHeader, SizeOf(BlockHeader))
  else
  begin
    FS_BlockRead(BIFFFile, BlockHeader, SizeOf(BlockHeader) - 4);
    BlockHeader.Unknown2 := 0
  end;

  case BlockHeader.TypeID of
    BlockType_Image:
      ProcessImage(BIFFFile, BlockHeader, IndLevel, ResourceText);
    BlockType_Palette:
      ProcessPalette(BIFFFile, BlockHeader, IndLevel, ResourceText);
    BlockType_Lister:
      ProcessLister(BIFFFile, BlockHeader, IndLevel, ResourceText);
    BlockType_Object:
      ProcessObject(BIFFFile, BlockHeader, IndLevel, ResourceText);
    BlockType_Material:
      ProcessMaterial(BIFFFile, BlockHeader, IndLevel, ResourceText)
  else
    ProcessUnknown(BIFFFile, BlockHeader, IndLevel, ResourceText)
  end;

  FS_BlockRead(BIFFFile, BlockTrailer, SizeOf(BlockTrailer));

  for i := 1 to BlockTrailer.NumSubBlocks do
    ProcessBlock(BIFFFile, BIFFHeader, IndLevel + 2, ResourceText);

  ResourceText.Add(Format('%-*s}', [IndLevel, '']))
end;

procedure ProcessBIFF(var BIFFFile: TFS_Handle; const BIFFHeader: TBIFFHeader;
  var ResourceText: TStringList; const BIFFCallback: TBIFFCallback);
var
  i: Longint;
begin
  for i := 1 to BIFFHeader.NumBlocks do
  begin
    ProcessBlock(BIFFFile, BIFFHeader, 0, ResourceText);
    BIFFCallback(i / BIFFHeader.NumBlocks)
  end
end;

function ConvertBIFF(const BIFFName: string; var ResourceText: TStringList;
  const BIFFCallback: TBIFFCallback): Boolean;
var
  BIFFFile  : TFS_Handle;
  BIFFHeader: TBIFFHeader;
begin
  Result := OpenBIFF(BIFFFile, BIFFName);
  if not Result then
    Exit;

  Result := ReadHeader(BIFFFile, BIFFHeader);
  if not Result then
  begin
    FS_CloseFile(BIFFFile);
    Exit
  end;

  ProcessBIFF(BIFFFile, BIFFHeader, ResourceText, BIFFCallback);

  Result := (ResourceText.Count > 0);

  FS_CloseFile(BIFFFile)
end;

end.
