{ File   : TM.pas
  Author : Deniz Oezmen
  Created: 2007-02-14, based on TM2BMP.dpr created 2007-02-13
  Changed: 2007-03-18

  The Longest Journey TM converter.
}
unit TM;

interface

uses
  Graphics;

type
  TTMCallback = procedure(const Frac: Double);

  TTMEntry    = record
                  Name  : string;
                  Bitmap: TBitmap
                end;

  TTMEntries  = array of TTMEntry;

function ConvertTM(const TMName: string; out TMEntries: TTMEntries;
  UseMIPMap: Boolean; const TMCallback: TTMCallback): Boolean;

implementation

uses
  TLJ_FS;

const
  TMHeader_ID       = 'BIFF';
  Block_BeginMarker = $f0f0f0f0;
  Block_EndMarker   = $0f0f0f0f;
  BlockType_Image   = $02faf080;
  BlockType_Palette = $02faf082;

type
  TTMHeader      = packed record
                     ID       : array[0..3] of Char;
                     Version,   // 1 or 2 allowed
                     Unknown1,
                     Unknown2,
                     NumBlocks: Longword;
                   end;

  TBlockHeader   = packed record
                     BeginMarker,
                     TypeID,
                     Unknown1   : Longword;
                     DataSize   : Longint;
                     Unknown2   : Longword; //only exists in version 2
                   end;

  TBlockTrailer  = packed record
                     EndMarker,
                     NumSubBlocks: Longword;
                   end;

  TPaletteHeader = packed record
                     NumEntries: Longword;
                   end;

  TImageHeader   = packed record
                     NameLen  : Word;
                     Name     : string;
                     Unknown1 : Byte;
                     Width,
                     Height   : Longint;
                     NumLevels: Longword;
                   end;

  TRGB           = packed record
                     R, G, B: Word;
                   end;

  TPalette       = array of TRGB;


function OpenTM(var TMFile: TFS_Handle; TMName: string): Boolean;
begin
  Result := FS_OpenFile(TMName, TMFile)
end;

function ReadHeader(var TMFile: TFS_Handle; out TMHeader: TTMHeader): Boolean;
var
  BytesRead: Longword;
begin
  BytesRead := FS_BlockRead(TMFile, TMHeader, SizeOf(TTMHeader));
  ReadHeader := (SizeOf(TTMHeader) = BytesRead) and
    (TMHeader.ID = TMHeader_ID) and (TMHeader.Version in [1, 2])
end;

procedure ReadPalette(var TMFile: TFS_Handle; out Palette: TPalette);
var
  PaletteHeader: TPaletteHeader;
begin
  FS_BlockRead(TMFile, PaletteHeader, SizeOf(PaletteHeader));
  SetLength(Palette, PaletteHeader.NumEntries);
  FS_BlockRead(TMFile, Palette[0], Length(Palette) * SizeOf(TRGB))
end;

procedure ConvertImage(var TMFile: TFS_Handle; const Palette: TPalette;
  var TMEntries: TTMEntries; UseMIPMap: Boolean);
var
  i          : Longint;
  ImageHeader: TImageHeader;
  ImageData  : array of Byte;
  Bitmap     : TBitmap;
  RGB        : TRGB;
  Pixel,
  Left,
  Top,
  x, y       : Longint;
begin
  with ImageHeader do
  begin
    FS_BlockRead(TMFile, NameLen, SizeOf(NameLen));
    SetLength(Name, NameLen);
    FS_BlockRead(TMFile, Name[1], NameLen);
    SetLength(Name, NameLen - 1);
    FS_BlockRead(TMFile, Unknown1, SizeOf(Unknown1));
    FS_BlockRead(TMFile, Width, SizeOf(Width));
    FS_BlockRead(TMFile, Height, SizeOf(Height));
    FS_BlockRead(TMFile, NumLevels, SizeOf(NumLevels));

    Bitmap := TBitmap.Create;
    if (NumLevels > 1) and UseMIPMap then
      Bitmap.Width := Width + Width shr 1
    else
      Bitmap.Width := Width;
    Bitmap.Height := Height;

    Left := 0;
    Top  := 0;
    for i := 1 to NumLevels do
    begin
      SetLength(ImageData, Width * Height);
      FS_BlockRead(TMFile, ImageData[0], Length(ImageData));

      if (i = 1) or UseMIPMap then
      begin
        for y := 0 to Height - 1 do
          for x := 0 to Width - 1 do
          begin
            Pixel := ImageData[y * Width + x];

            if Pixel > High(Palette) then
            begin
              RGB.R := 0;
              RGB.G := 0;
              RGB.B := 0
            end
            else
              with Palette[Pixel] do
                Bitmap.Canvas.Pixels[Left + x, Top + y] :=
                  B shl 16 or G shl 8 or R
          end;

        if Left = 0 then
          Left := Width
        else
          Top := Top + Height
      end;

      Width := Width shr 1;
      Height := Height shr 1
    end;

    SetLength(TMEntries, Length(TMEntries) + 1);
    TMEntries[High(TMEntries)].Name := Name;
    TMEntries[High(TMEntries)].Bitmap := Bitmap
  end;

  Finalize(ImageData)
end;

procedure ProcessBlock(var TMFile: TFS_Handle; const TMHeader: TTMHeader;
  Palette: TPalette; var TMEntries: TTMEntries; UseMIPMap: Boolean);
var
  i           : Longint;
  BlockHeader : TBlockHeader;
  BlockTrailer: TBlockTrailer;
begin
  if TMHeader.Version = 2 then
    FS_BlockRead(TMFile, BlockHeader, SizeOf(BlockHeader))
  else
  begin
    FS_BlockRead(TMFile, BlockHeader, SizeOf(BlockHeader) - 4);
    BlockHeader.Unknown2 := 0
  end;

  case BlockHeader.TypeID of
    BlockType_Image:
      ConvertImage(TMFile, Palette, TMEntries, UseMIPMap);
    BlockType_Palette:
      ReadPalette(TMFile, Palette)
  else
    FS_Seek(TMFile, FS_FilePos(TMFile) + BlockHeader.DataSize)
  end;

  FS_BlockRead(TMFile, BlockTrailer, SizeOf(BlockTrailer));

  for i := 1 to BlockTrailer.NumSubBlocks do
    ProcessBlock(TMFile, TMHeader, Palette, TMEntries, UseMIPMap);

  Finalize(Palette)
end;

procedure ProcessTM(var TMFile: TFS_Handle; const TMHeader: TTMHeader;
  out TMEntries: TTMEntries; UseMIPMap: Boolean; const TMCallback: TTMCallback);
var
  i      : Longint;
  Palette: TPalette;
begin
  SetLength(Palette, 0);
  SetLength(TMEntries, 0);

  for i := 1 to TMHeader.NumBlocks do
  begin
    ProcessBlock(TMFile, TMHeader, Palette, TMEntries, UseMIPMap);
    TMCallback(i / TMHeader.NumBlocks)
  end
end;

function ConvertTM(const TMName: string; out TMEntries: TTMEntries;
  UseMIPMap: Boolean; const TMCallback: TTMCallback): Boolean;
var
  TMFile  : TFS_Handle;
  TMHeader: TTMHeader;
begin
  Result := OpenTM(TMFile, TMName);
  if not Result then
    Exit;

  Result := ReadHeader(TMFile, TMHeader);
  if not Result then
  begin
    FS_CloseFile(TMFile);
    Exit
  end;

  ProcessTM(TMFile, TMHeader, TMEntries, UseMIPMap, TMCallback);

  Result := (Length(TMEntries) > 0);

  FS_CloseFile(TMFile)
end;

end.
