{ File   : XMG.pas
  Author : Deniz Oezmen
  Created: 2006-09-01, based on XMG2BMP.dpr created 2005-11-01
  Changed: 2010-02-24

  The Longest Journey XMG converter.
}
unit XMG;

interface

uses
  Graphics;

type
  TXMGCallback = procedure(const Frac: Double);

function ConvertXMG(const XMGName: string; out Bitmap: TBitmap;
  const XMGCallback: TXMGCallback): Boolean;

implementation

uses
  TLJ_FS;

type
  TYUVEntry = packed record
                Y : array[0..3] of Byte;
                Cr,
                Cb: Byte
              end;

  TYUVTable = array[0..255] of Longint;

  TRGB      = packed record
                R, G, B: Byte
              end;

  TRGBEntry = array[0..3] of TRGB;

  TXMGHeader = record
                 Version : Longword;
                 VoidClr : TRGB;
                 Empty   : Byte;
                 Width,
                 Height,
                 LineLen,
                 Unknown3,
                 Unknown4: Longword
               end;

var
  YCrR,
  YCrG,
  YCbG,
  YCbB,
  YTable: TYUVTable;

procedure SetupYUVTables;
var
  i        : Longint;
  x,
  YCbBScale,
  YCrRScale,
  YCbGScale,
  YCrGScale: Double;
begin
(* general YUV (better: YCbCr) to RGB computation scheme:

   | R |   |  1.000000  1.402000  0.000000 |   | Y  |
   | G | = |  1.000000 -0.714136 -0.344136 | * | Cr |
   | B |   |  1.000000  0.000000  1.772000 |   | Cb |

  The following precomputed lookup tables are approximative due to rounding.
  Compared to the game-generated pictures, small differences (usually
  unnoticable) are possible.
 *)
	YCbBScale :=  1.772000 * 256;
	YCrRScale :=  1.402000 * 256;
	YCbGScale := -0.344136 * 256;
	YCrGScale := -0.714136 * 256;

  for i := 0 to 255 do
  begin
    x := i - 128;

		YCbB[i] := Trunc(YCbBScale * x);
		YCrR[i] := Trunc(YCrRScale * x);
		YCbG[i] := Trunc(YCbGScale * x);
		YCrG[i] := Trunc(YCrGScale * x);
		YTable[i] := i shl 8
  end
end;

function OpenXMG(var XMGFile: TFS_Handle; XMGName: string): Boolean;
begin
  Result := FS_OpenFile(XMGName, XMGFile)
end;

function ReadHeader(var XMGFile: TFS_Handle;
  out XMGHeader: TXMGHeader): Boolean;
var
  BytesRead: Longint;
begin
  BytesRead := FS_BlockRead(XMGFile, XMGHeader, SizeOf(TXMGHeader));
  ReadHeader := (SizeOf(TXMGHeader) = BytesRead) and (XMGHeader.Version = 3)
    and (XMGHeader.Width * 3 = XMGHeader.LineLen)
end;

function Clamp(i: Longint; Max: Longint): Longint;
begin
  if i < 0 then
    Result := 0
  else if i > Max then
    Result := Max
  else
    Result := i
end;

procedure YUVToRGB(YUVEntry: TYUVEntry; var RGBEntry: TRGBEntry);
var
  i : Byte;
  YY: Longint;
begin
  for i := 0 to 3 do
    with RGBEntry[i] do
      with YUVEntry do
      begin
        YY := YTable[Y[i]];
        R := Clamp(YY + YCrR[Cr], 65535) shr 8;
        G := Clamp(YY + YCbG[Cb] + YCrG[Cr], 65535) shr 8;
        B := Clamp(YY + YCbB[Cb], 65535) shr 8
      end
end;

procedure Convert(var XMGFile: TFS_Handle; const XMGHeader: TXMGHeader;
  out Bitmap: TBitmap; const XMGCallback: TXMGCallback);
const
  cYUV  = 0;
  cVoid = 1;
  cRGB  = 2;
var
  i,
  BufPos,
  RunLen   : Longint;
  Opcode,
  ColorMode: Byte;
  x, y     : Longword;
  YUVEntry : TYUVEntry;
  Entries  : array of TRGBEntry;
  Buffer   : array of Byte;
begin
  Bitmap := TBitmap.Create;
  Bitmap.Width := XMGHeader.Width;
  Bitmap.Height := XMGHeader.Height;
  Bitmap.Transparent := True;

  with XMGHeader.VoidClr do
    Bitmap.TransparentColor := B shl 16 or G shl 8 or R;

  y := 0;
  x := 0;
  RunLen := 0;
  ColorMode := cVoid;

  SetLength(Buffer, FS_FileSize(XMGFile));
  FS_BlockRead(XMGFile, Buffer[0], FS_FileSize(XMGFile));
  BufPos := 0;

  XMGCallback(0);

  while BufPos < Length(Buffer) do
  begin
    OpCode := Buffer[BufPos];
    Inc(BufPos);

    (* determine entry type. Each opcode is interpreted as a bitfield of the
       format xxyyzzzz, where xx denotes the color mode (00, 01, 10) and yyzzzz
       denotes the number of 2x2 pixel blocks to be processed.
       In case xx equals 11, yy is the color mode and a second opcode aaaaaaaa
       has to be read. The number of blocks is then zzzzaaaaaaaa.
     *)
    case OpCode of
      $00..$3f: begin
                  ColorMode := cYUV;
                  RunLen := OpCode
                end;
      $40..$7f: begin
                  ColorMode := cVoid;
                  RunLen := OpCode and $3f
                end;
      $80..$bf: begin
                  ColorMode := cRGB;
                  RunLen := OpCode and $3f
                end;
      $c0..$fe: begin
                  // need to read further parameter
                  ColorMode := (OpCode shr 4) and $03;
                  RunLen := (OpCode and $0f) shl 8;

                  OpCode := Buffer[BufPos];
                  Inc(BufPos);
                  RunLen := RunLen + OpCode
                end;
           $ff: Break
    end;

    // read entries
    SetLength(Entries, RunLen);
    case ColorMode of
      cVoid: for i := 0 to RunLen - 1 do
             begin
               Entries[i][0] := XMGHeader.VoidClr;
               Entries[i][1] := XMGHeader.VoidClr;
               Entries[i][2] := XMGHeader.VoidClr;
               Entries[i][3] := XMGHeader.VoidClr
             end;
      cYUV : for i := 0 to RunLen - 1 do
             begin
               Move(Buffer[BufPos], YUVEntry, SizeOf(TYUVEntry));
               Inc(BufPos, SizeOf(TYUVEntry));
               YUVToRGB(YUVEntry, Entries[i])
             end;
      cRGB : begin
               Move(Buffer[BufPos], Entries[0], SizeOf(TRGBEntry) * RunLen);
               Inc(BufPos, SizeOf(TRGBEntry) * RunLen)
             end
    end;

    // draw entries
    for i := 0 to RunLen - 1 do
    begin
      with Entries[i][0] do
        Bitmap.Canvas.Pixels[x    , y    ] := B shl 16 or G shl 8 or R;
      with Entries[i][1] do
        Bitmap.Canvas.Pixels[x + 1, y    ] := B shl 16 or G shl 8 or R;
      with Entries[i][2] do
        Bitmap.Canvas.Pixels[x    , y + 1] := B shl 16 or G shl 8 or R;
      with Entries[i][3] do
        Bitmap.Canvas.Pixels[x + 1, y + 1] := B shl 16 or G shl 8 or R;

      Inc(x, 2);

      if x >= XMGHeader.Width then
      begin
        x := 0;
        Inc(y, 2)
      end
    end;

    XMGCallback(BufPos / Length(Buffer))
  end;

  XMGCallback(1);

  Finalize(Entries)
end;

function ConvertXMG(const XMGName: string; out Bitmap: TBitmap;
  const XMGCallback: TXMGCallback): Boolean;
var
  XMGFile  : TFS_Handle;
  XMGHeader: TXMGHeader;
begin
  Result := OpenXMG(XMGFile, XMGName);
  if not Result then
    Exit;

  Result := ReadHeader(XMGFile, XMGHeader);
  if not Result then
  begin
    FS_CloseFile(XMGFile);
    Exit
  end;

  Convert(XMGFile, XMGHeader, Bitmap, XMGCallback);

  FS_CloseFile(XMGFile)
end;

initialization
  SetupYUVTables;

end.

