{ File   : TLJ_FS.pas
  Author : Deniz Oezmen
  Created: 2006-09-01
  Changed: 2007-08-04

  Naive file system abstraction layer for The Longest Journey.
}
unit TLJ_FS;

interface

type
  TTLJ_FS_Callback = procedure(const Path: string);

type
  TFS_Handle  = record
                  FS_File: file;
                  BaseOfs,
                  Size   : Longint;
                end;

  TFS_Entry   = record
                  Name,
                  FriendlyName: string;
                  ExtendedInfo: array of string;
                end;

  TFS_Entries = array of TFS_Entry;

function FS_Init(const Dir: string; Callback: TTLJ_FS_Callback): Boolean;
procedure FS_DeInit;
function FS_ChangeDir(const Dir: string): Boolean;
function FS_GetCurrentDir: string;
procedure FS_GetFileList(out FileList: TFS_Entries); overload;
procedure FS_GetDirList(out DirList: TFS_Entries); overload;
function FS_GetInfo(const Path: string; out FS_Entry: TFS_Entry): Boolean;
function FS_OpenFile(const FileName: string;
  out FS_Handle: TFS_Handle): Boolean;
procedure FS_CloseFile(var FS_Handle: TFS_Handle);
function FS_FilePos(var FS_Handle: TFS_Handle): Longint;
function FS_FileSize(var FS_Handle: TFS_Handle): Longint;
function FS_BlockRead(var FS_Handle: TFS_Handle; var Buffer;
  Count: Longint): Longint;
procedure FS_Seek(var FS_Handle: TFS_Handle; RelPos: Longint);
function FS_Eof(var FS_Handle: TFS_Handle): Boolean;

implementation

uses
  XARC_FS, XRC_FS, QSort,
  SysUtils;

const
  FS_NodeType_Root      = 1;
  FS_NodeType_Directory = 2;
  FS_NodeType_File      = 4;
  FS_NodeType_InArchive = 8;

  // files larger than MaxFileSize are considered part of the copy protection
  // scheme of the 4 CD version and will be omitted during scan
  MaxFileSize = 838860800;

type
  PFS_Node = ^TFS_Node;
  TFS_Node = record
               NodeType    : Longint;
               Name,
               FriendlyName: string;
               ExtendedInfo: array of string;
               Offset,
               Size        : Longint;
               Parent      : PFS_Node;
               Children    : array of PFS_Node;
             end;

var
  RootDir: string;
  FS_Root,
  FS_CWD : PFS_Node;

// forward declarations for internal methods

procedure Traverse(const FS_Node: PFS_Node; const Dir: string;
  Callback: TTLJ_FS_Callback); forward;
function FindNode(Name: string; CurNode: PFS_Node; out Node: PFS_Node): Boolean;
  forward;
procedure DestroyNode(var FS_Node: PFS_Node); forward;
function AddToFS(RelDir, Dir: string; Callback: TTLJ_FS_Callback): Boolean;
  forward;
function GetFullName(FS_Node: PFS_Node): string; forward;
function FS_FileEntryIsLess(var DataA, DataB): Boolean; forward;
function FS_DirEntryIsLess(var DataA, DataB): Boolean; forward;

(* published methods start here ***********************************************)

function FS_Init(const Dir: string; Callback: TTLJ_FS_Callback): Boolean;
begin
  Result := False;

  FS_DeInit;

  RootDir := Dir;
  if RootDir <> '' then
    while RootDir[Length(RootDir)] = '\' do
      RootDir := Copy(RootDir, 1, Length(RootDir) - 1);

  New(FS_Root);
  with FS_Root^ do
  begin
    NodeType := FS_NodeType_Root or FS_NodeType_Directory;
    Name := 'x';
    FriendlyName := '';
    Offset := -1;
    Size := -1;
    Parent := FS_Root;
    SetLength(Children, 0)
  end;

  Traverse(FS_Root, '\', Callback);

  // scan succeeded?
  if FS_Root^.FriendlyName <> 'LAIDBACK' then
  begin
    FS_DeInit;
    Exit
  end;

  // correct curious "LAIDBACK" friendly name for FS root into something
  // prettier
  FS_Root^.FriendlyName := 'The Longest Journey';

  // add special folders not found in main structure
  AddToFS('\', 'static\', Callback);
  AddToFS('\static\', 'diaryfmv\', Callback);
  AddToFS('\static\', 'diaryindexlocation\', Callback);
  AddToFS('\static\', 'diarylog\', Callback);
  AddToFS('\static\', 'diarypages\', Callback);
  AddToFS('\static\', 'loadsavelocation\', Callback);
  AddToFS('\static\', 'mainmenulocation\', Callback);
  AddToFS('\static\', 'optionlocation\', Callback);

  // current directory is root
  FS_CWD := FS_Root;

  Result := True
end;

procedure FS_DeInit;
begin
  DestroyNode(FS_Root);

  RootDir := ''
end;

function FS_ChangeDir(const Dir: string): Boolean;
var
  FS_Node: PFS_Node;
begin
  Result := False;

  if not FindNode(Dir, FS_CWD, FS_Node) then
    Exit;

  if FS_Node^.NodeType and FS_NodeType_Directory = 0 then
    Exit;

  FS_CWD := FS_Node;
  Result := True
end;

function FS_GetCurrentDir: string;
begin
  Result := GetFullName(FS_CWD)
end;

procedure FS_GetFileList(out FileList: TFS_Entries);
var
  i,
  Ctr : Longint;
begin
  if FS_CWD = nil then
    Exit;

  Ctr := 0;

  with FS_CWD^ do
    for i := Low(Children) to High(Children) do
      if Children[i]^.NodeType and FS_NodeType_File <> 0 then
        Inc(Ctr);

  SetLength(FileList, Ctr);

  Ctr := 0;
  with FS_CWD^ do
    for i := Low(Children) to High(Children) do
      if Children[i]^.NodeType and FS_NodeType_File <> 0 then
      begin
        FileList[Ctr].Name := Children[i]^.Name;
        FileList[Ctr].FriendlyName := Children[i]^.FriendlyName;
        Inc(Ctr)
      end;

  QuickSort(FileList, SizeOf(TFS_Entry), FS_FileEntryIsLess)
end;

procedure FS_GetDirList(out DirList: TFS_Entries);
var
  i,
  Ctr : Longint;
begin
  if FS_CWD = nil then
    Exit;

  Ctr := 0;

  with FS_CWD^ do
    for i := Low(Children) to High(Children) do
      if Children[i]^.NodeType and FS_NodeType_Directory <> 0 then
        Inc(Ctr);

  SetLength(DirList, Ctr);

  Ctr := 0;
  with FS_CWD^ do
    for i := Low(Children) to High(Children) do
      if Children[i]^.NodeType and FS_NodeType_Directory <> 0 then
      begin
        DirList[Ctr].Name := Children[i]^.Name;
        DirList[Ctr].FriendlyName := Children[i]^.FriendlyName;
        Inc(Ctr)
      end;

  QuickSort(DirList, SizeOf(TFS_Entry), FS_DirEntryIsLess)
end;

function FS_GetInfo(const Path: string; out FS_Entry: TFS_Entry): Boolean;
var
  i      : Longint;
  FS_Node: PFS_Node;
begin
  Result := FindNode(Path, FS_CWD, FS_Node);
  if not Result then
    Exit;

  with FS_Entry do
  begin
    Name := FS_Node^.Name;
    FriendlyName := FS_Node^.FriendlyName;
    SetLength(ExtendedInfo, Length(FS_Node^.ExtendedInfo));
    for i := 0 to High(ExtendedInfo) do
      ExtendedInfo[i] := FS_Node^.ExtendedInfo[i]
  end
end;

function FS_OpenFile(const FileName: string;
  out FS_Handle: TFS_Handle): Boolean;
var
  FS_Node: PFS_Node;
  Dir    : string;
begin
  // find file
  Result := FindNode(FileName, FS_CWD, FS_Node);
  if not Result then
    Exit;

  // is this really a file?
  Result := (FS_Node^.NodeType and FS_NodeType_File <> 0);
  if not Result then
    Exit;

  // get physical directory name
  Dir := RootDir + GetFullName(FS_Node^.Parent);

  FileMode := fmOpenRead;

  // archive file or regular file?
  if FS_Node^.NodeType and FS_NodeType_InArchive <> 0 then
    AssignFile(FS_Handle.FS_File, Dir + FS_Node^.Parent^.Name + '.xarc')
  else
    AssignFile(FS_Handle.FS_File, Dir + 'xarc\' + FS_Node^.Name);

  with FS_Handle do
  begin
    Reset(FS_File, 1);
    BaseOfs := FS_Node^.Offset;
    Size := FS_Node^.Size;
    Seek(FS_File, BaseOfs)
  end
end;

procedure FS_CloseFile(var FS_Handle: TFS_Handle);
begin
  CloseFile(FS_Handle.FS_File)
end;

function FS_FilePos(var FS_Handle: TFS_Handle): Longint;
begin
  with FS_Handle do
    Result := FilePos(FS_File) - BaseOfs
end;

function FS_FileSize(var FS_Handle: TFS_Handle): Longint;
begin
  with FS_Handle do
    Result := Size
end;

function FS_BlockRead(var FS_Handle: TFS_Handle; var Buffer;
  Count: Longint): Longint;
var
  RelPos: Longint;
begin
  with FS_Handle do
  begin
    RelPos := FS_FilePos(FS_Handle);

    if RelPos + Count > Size then
      Count := Size - RelPos;

    BlockRead(FS_File, Buffer, Count, Result)
  end
end;

procedure FS_Seek(var FS_Handle: TFS_Handle; RelPos: Longint);
var
  AbsPos: Longint;
begin
  with FS_Handle do
  begin
    if RelPos > Size then
      RelPos := Size;

    AbsPos := BaseOfs + RelPos;

    Seek(FS_File, AbsPos)
  end
end;

function FS_Eof(var FS_Handle: TFS_Handle): Boolean;
begin
  with FS_Handle do
    Result := (FS_FilePos(FS_Handle) = Size)
end;

(* Internal methods start here ************************************************)

function FindNode(Name: string; CurNode: PFS_Node; out Node: PFS_Node): Boolean;
var
  i: Longint;
  s: string;
begin
  Result := False;

  if CurNode = nil then
    Exit;

  // check for absolute path expression
  if Copy(Name, 1, 1) = '\' then
    CurNode := FS_Root;

  while Length(Name) > 0 do
  begin
    // cut multiple backslashes
    while Copy(Name, 1, 1) = '\' do
      Name := Copy(Name, 2, Length(Name) - 1);

    if Length(Name) = 0 then
      Break;

    // find next subpath
    i := Pos('\', Name);
    if i > 0 then
      s := Copy(Name, 1, i - 1)
    else
      s := Name;

    // process subpath
    if s = '..' then
      CurNode := CurNode^.Parent
    else if s <> '.' then
    begin
      // search for subpath
      with CurNode^ do
      begin
        for i := Low(Children) to High(Children) do
          if AnsiLowerCase(Children[i].Name) = AnsiLowerCase(s) then
          begin
            CurNode := Children[i];
            Break
          end;
        // failure
        if i = High(Children) + 1 then
          Exit
      end
    end;

    // cut subpath from expression
    Name := Copy(Name, Length(s) + 1, Length(Name) - Length(s))
  end;

  Node := CurNode;
  Result := True
end;

procedure DestroyNode(var FS_Node: PFS_Node);
var
  i: Longint;
begin
  if FS_Node = nil then
    Exit;

  with FS_Node^ do
  begin
    Finalize(Name);
    Finalize(FriendlyName);
    Parent := nil;
    for i := Low(Children) to High(Children) do
      DestroyNode(Children[i]);
    Finalize(Children)
  end;

  Dispose(FS_Node);
  FS_Node := nil
end;

function AddToFS(RelDir, Dir: string; Callback: TTLJ_FS_Callback): Boolean;
var
  Idx     : Longint;
  BaseNode,
  FS_Node : PFS_Node;
  Mtch    : TFilenameCaseMatch;
begin
  // find base directory node
  Result := FindNode(RelDir, FS_Root, BaseNode);
  if not Result then
    Exit;

  // cut trailing backslashes of subpath
  while Copy(Dir, Length(Dir), 1) = '\' do
    Dir := Copy(Dir, 1, Length(Dir) - 1);

  // absolute path or path contains multiple subpaths
  if Pos('\', Dir) > 0 then
    Exit;

  // test for subdirectory
  if not DirectoryExists(RootDir + RelDir + Dir) then
    Exit;

  // add new node
  New(FS_Node);
  with FS_Node^ do
  begin
    NodeType := FS_NodeType_Directory;
    Name := Dir;
    Offset := -1;
    Size := -1;
    Parent := BaseNode;
    SetLength(Children, 0)
  end;

  Idx := Length(BaseNode^.Children);
  SetLength(BaseNode^.Children, Idx + 1);
  BaseNode^.Children[Idx] := FS_Node;

  // correct case of subnode name
  FS_Node^.Name := ExtractFileName(ExpandFileNameCase(
    RootDir + RelDir + FS_Node^.Name, Mtch));

  // recurse
  Traverse(FS_Node, RelDir + Dir + '\', Callback);

  Result := True
end;

procedure Traverse(const FS_Node: PFS_Node; const Dir: string;
  Callback: TTLJ_FS_Callback);
var
  i, j, k,
  Idx    : Longint;
  XARC   : TXARC;
  XRC    : TXRC;
  SR     : TSearchRec;
  Mtch   : TFilenameCaseMatch;
  Ext,
  XRCName: string;
begin
  if FS_Node = nil then
    Exit;

  // first find matching XARC archive
  XARC.Name := Format('%s%s%s.xarc', [RootDir, Dir, FS_Node^.Name]);
  Callback(XARC.Name);
  if not ReadXARC(XARC) then
  begin
    ClearXARC(XARC);
    Exit
  end;

  XRCName := AnsiLowerCase(FS_Node^.Name) + '.xrc';

  // add XARC entries to Children list
  for i := Low(XARC.Entries) to High(XARC.Entries) do
    if AnsiLowerCase(XARC.Entries[i].Name) = XRCName then
    begin
      // resource entry
      XRC.Name := XARC.Name;
      XRC.BaseOfs := XARC.Entries[i].Offset;
      XRC.Size := XARC.Entries[i].Size;
      ReadXRC(XRC);

      FS_Node^.FriendlyName := XRC.Header.Name;

      // get sublocations
      Idx := Length(FS_Node^.Children);
      SetLength(FS_Node^.Children,
        Length(FS_Node^.Children) + Length(XRC.Locations));
      for j := Low(XRC.Locations) to High(XRC.Locations) do
      begin
        New(FS_Node^.Children[Idx]);
        with FS_Node^.Children[Idx]^ do
        begin
          NodeType := FS_NodeType_Directory;
          FriendlyName := XRC.Locations[j].Name;
          Name := '';
          Offset := -1;
          Size := -1;
          Parent := FS_Node;
          SetLength(Children, 0)
        end;

        // test for subdirectories
        if DirectoryExists(RootDir + Dir + FS_Node^.Children[Idx]^.FriendlyName)
          then
          FS_Node^.Children[Idx]^.Name := FS_Node^.Children[Idx]^.FriendlyName
        else if DirectoryExists(RootDir + Dir + FS_Node^.Children[Idx]^.Name)
          then
          FS_Node^.Children[Idx]^.Name := Format('%.2x', [XRC.Locations[j].ID])
        else
          Continue;

        // correct case of subnode name
        FS_Node^.Children[Idx]^.Name := ExtractFileName(ExpandFileNameCase(
          RootDir + Dir + FS_Node^.Children[Idx]^.Name, Mtch));

        // recurse
        Traverse(FS_Node^.Children[Idx],
          Dir + FS_Node^.Children[Idx]^.Name + '\', Callback);

        Inc(Idx)
      end
    end;

  // count standard file entries
  Idx := Length(XARC.Entries);

  SetLength(FS_Node^.Children, Length(FS_Node^.Children) + Idx);
  Idx := Length(FS_Node^.Children) - Idx;

  // add standard file entries to Children list
  for i := Low(XARC.Entries) to High(XARC.Entries) do
  begin
    // standard file entry
    New(FS_Node^.Children[Idx]);
    with FS_Node^.Children[Idx]^ do
    begin
      NodeType := FS_NodeType_File or FS_NodeType_InArchive;
      Name := XARC.Entries[i].Name;
      FriendlyName := '';
      Offset := XARC.Entries[i].Offset;
      Size := XARC.Entries[i].Size;
      Parent := FS_Node;
      SetLength(Children, 0)
    end;
    Inc(Idx)
  end;

  // now scan physical directory for additional files
  Callback(Format('%s%sxarc', [RootDir, Dir]));
  if FindFirst(Format('%s%sxarc\*.*', [RootDir, Dir]),
    faAnyFile - faDirectory, SR) = 0 then
  repeat
    // omit copy protection files
    if SR.Size > MaxFileSize then
      Continue;

    Idx := Length(FS_Node^.Children);
    SetLength(FS_Node^.Children, Idx + 1);
    New(FS_Node^.Children[Idx]);
    with FS_Node^.Children[Idx]^ do
    begin
      NodeType := FS_NodeType_File;
      Name := SR.Name;
      FriendlyName := '';
      Offset := 0;
      Size := SR.Size;
      Parent := FS_Node;
      SetLength(Children, 0)
    end
  until FindNext(SR) <> 0;

  FindClose(SR);

  // search for XRC friendly names
  for i := Low(XRC.Files) to High(XRC.Files) do
    for j := Low(FS_Node^.Children) to High(FS_Node^.Children) do
      with FS_Node^.Children[j]^ do
      begin
        if FriendlyName <> '' then
          Continue;

        if AnsiLowerCase(XRC.Files[i].Name) = AnsiLowerCase(Name) then
        begin
          FriendlyName := XRC.Files[i].FriendlyName;
          SetLength(ExtendedInfo, Length(XRC.Files[i].ExtendedInfo));
          for k := 0 to High(ExtendedInfo) do
            ExtendedInfo[k] := XRC.Files[i].ExtendedInfo[k]
        end;

        // OVS fix (files are still referred to as .isn/... files internally)
        Ext := AnsiLowerCase(ExtractFileExt(XRC.Files[i].Name));
        if (Ext = '.isn') or (Ext = '.iss') or (Ext = '.ssn') or
          (Ext = '.sn') then
          if AnsiLowerCase(ChangeFileExt(XRC.Files[i].Name, '.ovs')) =
            AnsiLowerCase(Name) then
          begin
            FriendlyName := XRC.Files[i].FriendlyName;
            SetLength(ExtendedInfo, Length(XRC.Files[i].ExtendedInfo));
            for k := 0 to High(ExtendedInfo) do
              ExtendedInfo[k] := XRC.Files[i].ExtendedInfo[k]
          end
      end;

  ClearXARC(XARC);
  ClearXRC(XRC)
end;

function GetFullName(FS_Node: PFS_Node): string;
begin
  Result := '';

  if FS_Node = nil then
    Exit;

  if FS_Node^.NodeType and FS_NodeType_Directory <> 0 then
    Result := '\';

  while FS_Node <> FS_Root do
  begin
    Result := '\' + FS_Node^.Name + Result;
    FS_Node := FS_Node^.Parent
  end
end;

function FS_FileEntryIsLess(var DataA, DataB): Boolean;
var
  FS_EntryA,
  FS_EntryB: TFS_Entry;
  CompA,
  CompB,
  ExtA,
  ExtB     : string;
begin
  FS_EntryA := TFS_Entry(DataA);
  FS_EntryB := TFS_Entry(DataB);

  CompA := AnsiLowerCase(FS_EntryA.Name);
  CompB := AnsiLowerCase(FS_EntryB.Name);

  ExtA := ExtractFileExt(CompA);
  ExtB := ExtractFileExt(CompB);

  // sorting criteria: extension first, name/friendly name afterwards
  if ExtA = ExtB then
  begin
    if FS_EntryA.FriendlyName <> '' then
      CompA := AnsiLowerCase(FS_EntryA.FriendlyName);
    if FS_EntryB.FriendlyName <> '' then
      CompB := AnsiLowerCase(FS_EntryB.FriendlyName);

    Result := (CompA < CompB)
  end
  else
    Result := (ExtA < ExtB)
end;

function FS_DirEntryIsLess(var DataA, DataB): Boolean;
var
  FS_EntryA,
  FS_EntryB: TFS_Entry;
  CompA,
  CompB    : string;

  function IsHexByte(s: string): Boolean;
  var
    i: Longint;
  begin
    Result := False;

    if not (Length(s) in [1, 2]) then
      Exit;

    Result := True;

    for i := 1 to Length(s) do
      Result := Result and (s[i] in ['0'..'9', 'a'..'f'])
  end;

begin
  FS_EntryA := TFS_Entry(DataA);
  FS_EntryB := TFS_Entry(DataB);

  CompA := AnsiLowerCase(FS_EntryA.Name);
  CompB := AnsiLowerCase(FS_EntryB.Name);

  // sorting criterion: alpha before hex, lexicographical otherwise
  if IsHexByte(CompA) and not IsHexByte(CompB) then
    Result := False
  else if not IsHexByte(CompA) and IsHexByte(CompB) then
    Result := True
  else
    Result := (CompA < CompB)
end;

initialization
  FS_Root := nil;
  RootDir := '';

finalization
  FS_DeInit;

end.
