{ File   : ExpHndl.pas
  Author : Deniz Oezmen
  Created: 2007-03-20, based on Handlers.pas created 2007-02-15
  Changed: 2010-02-24

  TLJView data export handler.
}
unit ExpHndl;

interface

uses
  HndlType;

const
  // export types
  Export_None     = -1;
  Export_Single   = 0;
  Export_Multiple = 1;

function Handle_GetExportDefaultName(const HandlerData: THandlerData): string;
function Handle_GetExportType(const HandlerData: THandlerData): Longint;
function Handle_Export(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;

implementation

uses
  SysUtils, Classes, Graphics,
  Raw, ANI, SaveTGA;

function Handle_GetExportDefaultName(const HandlerData: THandlerData): string;
var
  Ext: string;
begin
  Result := '';

  if HandlerData.Name = '' then
    Exit;

  Ext := '';

  case HandlerData.HandlerType of
    Type_RawFile:
      Ext := ExtractFileExt(HandlerData.Name);
    Type_XMGPicture:
      Ext := '.tga';
    Type_OVSSound:
      Ext := '.ogg';
    Type_ISNSound:
      Ext := '.wav';
    Type_TMTexture:
      Ext := '.tga';
    Type_XRCResource:
      Ext := '.txt';
    Type_SSSAnimation:
      Ext := '.smk';
    Type_BBBAnimation:
      Ext := '.bik';
    Type_ANIAnimation:
      Ext := '.txt';
    Type_CIRModel:
      Ext := '.txt';
    Type_BIFFObject:
      Ext := '.txt'
  end;

  if Ext <> '' then
    Result := ChangeFileExt(ExtractFileName(HandlerData.Name), Ext)
end;

function Handle_GetExportType(const HandlerData: THandlerData): Longint;
begin
  case HandlerData.HandlerType of
    Type_Root:
      Result := Export_None;
    Type_FolderClose:
      Result := Export_None;
    Type_FolderOpen:
      Result := Export_None;
    Type_RawFile:
      Result := Export_Single;
    Type_XMGPicture:
      Result := Export_Single;
    Type_OVSSound:
      Result := Export_Single;
    Type_ISNSound:
      Result := Export_Single;
    Type_TMTexture:
      Result := Export_Multiple;
    Type_XRCResource:
      Result := Export_Single;
    Type_SSSAnimation:
      Result := Export_Single;
    Type_BBBAnimation:
      Result := Export_Single;
    Type_CIRModel:
      Result := Export_Single;
    Type_ANIAnimation:
      Result := Export_Single;
    Type_BIFFObject:
      Result := Export_Single
  else
    Result := Export_None
  end;
end;

function CopyFile(const SourceName, DestName: string;
  const HandleCallback: THandleCallback): Boolean;
const
  BufSize = 1048576;
var
  SrcFile,
  DstFile     : file;
  Buffer      : array of Byte;
  BytesRead,
  BytesWritten: Longint;
begin
  Result := False;

  FileMode := fmOpenRead;
  AssignFile(SrcFile, SourceName);
  {$I-}
  Reset(SrcFile, 1);
  {$I+}
  if IOResult <> 0 then
    Exit;

  FileMode := fmOpenWrite;
  AssignFile(DstFile, DestName);
  {$I-}
  Rewrite(DstFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    CloseFile(SrcFile);
    Exit
  end;

  SetLength(Buffer, BufSize);

  while not Eof(SrcFile) do
  begin
    BlockRead(SrcFile, Buffer[0], BufSize, BytesRead);
    if (BytesRead < BufSize) and not Eof(SrcFile) then
    begin
      Finalize(Buffer);
      CloseFile(SrcFile);
      CloseFile(DstFile);
      Exit
    end;

    BlockWrite(DstFile, Buffer[0], BytesRead, BytesWritten);
    if BytesWritten < BytesRead then
    begin
      Finalize(Buffer);
      CloseFile(SrcFile);
      CloseFile(DstFile);
      Exit
    end;

    HandleCallback(FilePos(SrcFile) / FileSize(SrcFile))
  end;

  Finalize(Buffer);
  CloseFile(SrcFile);
  CloseFile(DstFile);

  Result := True
end;

function StoreTGA(const Bitmap: TBitmap; const ExportName: string;
  const HandleCallback: THandleCallback): Boolean;
var
   ImageData   : array of Longword;
   ImageDataPtr,
   ScanLinePtr : PLongword;
   x, y        : Longint;
   Color,
   TransColor  : Longword;
begin
  Result := False;

  if not Assigned(Bitmap) then
    Exit;

  with Bitmap do
  begin
    SetLength(ImageData, Width * Height);
    ImageDataPtr := @ImageData[0];

    // switch transparent color component order for scanline compatibility
    TransColor := (TransparentColor and $0000ff00) or
                  (TransparentColor and $00ff0000) shr 16 or
                  (TransparentColor and $000000ff) shl 16;

    for y := 0 to Height - 1 do
    begin
      ScanLinePtr := ScanLine[y];

      for x := 0 to Width - 1 do
      begin
        Color := ScanLinePtr^;
        Inc(ScanLinePtr);

        // determine transparency
        if not Transparent or (Color <> TransColor) then
          Color := Color or $ff000000;

        ImageDataPtr^ := Color;
        Inc(ImageDataPtr)
      end;

      HandleCallback(y / Height)
    end;

    // save to file
    DumpTGA(ExportName, Width, Height, @ImageData[0],
      TGAPixelDepth_32Bit, TGACompression_RLE);

    SetLength(ImageData, 0)
  end;

  Result := True
end;

function Export_Bitmap(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  if Length(HandlerData.BitmapList) <= 0 then
    Exit;

  try
    Result := StoreTGA(HandlerData.BitmapList[0], ExportName, HandleCallback)
  except
    on Exception do
      ;
  end
end;

function Export_BitmapList(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
var
  i                : Longint;
  SingleHandlerData: THandlerData;
  SingleExportName : string;
begin
  Result := False;

  for i := 0 to High(HandlerData.BitmapList) do
    try
      SingleHandlerData.Name := HandlerData.StringList[i];
      SingleHandlerData.HandlerType := HandlerData.HandlerType;
      SingleHandlerData.Intf_Type := HandlerData.Intf_Type;
      SetLength(SingleHandlerData.BitmapList, 1);
      SingleHandlerData.BitmapList[0] := HandlerData.BitmapList[i];

      SingleExportName := ExportName + '\' +
        Handle_GetExportDefaultName(SingleHandlerData);

      if not Export_Bitmap(SingleHandlerData, SingleExportName, HandleCallback)
        then
        Exit
    except
      on Exception do
        Exit
    end;

  Result := True
end;

function Export_Sound(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  if HandlerData.SoundList.Count <= 0 then
    Exit;

  Result := CopyFile(HandlerData.SoundList[0], ExportName, HandleCallback)
end;

function Export_SoundComment(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  if HandlerData.SoundList.Count <= 0 then
    Exit;

  Result := CopyFile(HandlerData.SoundList[0], ExportName, HandleCallback)
end;

function Export_TextDataList(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  try
    HandlerData.StringList.SaveToFile(ExportName);
    Result := True
  except
    on Exception do
      ;
  end
end;

function Export_RawData(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
var
  DstFile     : file;
  Buffer      : TRawBuffer;
  BytesWritten: Longint;
begin
  Result := LoadRaw(HandlerData.Name, Buffer, HandleCallback);
  if not Result then
    Exit;

  FileMode := fmOpenWrite;
  AssignFile(DstFile, ExportName);
  {$I-}
  Rewrite(DstFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    Finalize(Buffer);
    Exit
  end;

  BlockWrite(DstFile, Buffer[0], Length(Buffer), BytesWritten);

  Result := (BytesWritten = Length(Buffer));

  Close(DstFile);
  Finalize(Buffer)
end;

function Export_ExtAnimation(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  if HandlerData.ExtAnimationList.Count <= 0 then
    Exit;

  Result := CopyFile(HandlerData.ExtAnimationList[0], ExportName,
    HandleCallback)
end;

function Export_IntAnimation(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
var
  ResourceText: TStringList;
begin
  ResourceText := TStringList.Create;

  Result := LoadANIAsText(HandlerData.Name, ResourceText, HandleCallback);
  if not Result then
    Exit;

  Result := False;

  try
    try
      ResourceText.SaveToFile(ExportName);
      Result := True
    finally
      ResourceText.Free
    end
  except
    on Exception do
      ;
  end
end;

function Handle_Export(const HandlerData: THandlerData;
  const ExportName: string; const HandleCallback: THandleCallback): Boolean;
begin
  Result := False;

  case Handle_GetExportType(HandlerData) of
    Export_Single:
      case HandlerData.Intf_Type of
        Intf_Bitmap:
          Result := Export_Bitmap(HandlerData, ExportName, HandleCallback);
        Intf_Sound:
          Result := Export_Sound(HandlerData, ExportName, HandleCallback);
        Intf_SoundComment:
          Result := Export_SoundComment(HandlerData, ExportName,
            HandleCallback);
        Intf_TextDataList:
          Result := Export_TextDataList(HandlerData, ExportName,
            HandleCallback);
        Intf_RawData:
          Result := Export_RawData(HandlerData, ExportName, HandleCallback);
        Intf_ExtAnimation:
          Result := Export_ExtAnimation(HandlerData, ExportName,
            HandleCallback);
        Intf_IntAnimation:
          Result := Export_IntAnimation(HandlerData, ExportName, HandleCallback)
      end;
    Export_Multiple:
      case HandlerData.Intf_Type of
        Intf_BitmapList:
          Result := Export_BitmapList(HandlerData, ExportName, HandleCallback);
      end
  end
end;

end.
