{ File   : XMG2BMP.dpr
  Author : Deniz Oezmen
  Created: 2005-11-01
  Changed: 2007-03-16

  The Longest Journey XMG converter.
}
program XMG2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils, Math, SaveBMP3;

type
  TYUVEntry = packed record
                Y : array[0..3] of Byte;
                Cr,
                Cb: Byte
              end;

  TYUVTable = array[0..255] of Longint;

  TRGB      = packed record
                R, G, B: Byte
              end;

  TRGBEntry = array[0..3] of TRGB;

  TXMGHeader = record
                 Version : Longword;
                 VoidClr : TRGB;
                 Empty   : Byte;
                 Width,
                 Height,
                 LineLen,
                 Unknown3,
                 Unknown4: Longword
               end;

var
  XMGFile  : file;
  XMGHeader: TXMGHeader;
  YCrR,
  YCrG,
  YCbG,
  YCbB,
  YTable   : TYUVTable;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenXMGFile(var XMGFile: file; XMGName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(XMGFile, XMGName);

  {$I-}
  Reset(XMGFile, 1);
  {$I+}
  OpenXMGFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(XMGFile, XMGHeader, SizeOf(TXMGHeader), BytesRead);
  ReadHeader := (SizeOf(TXMGHeader) = BytesRead) and (XMGHeader.Version = 3)
    and (XMGHeader.Width * 3 = XMGHeader.LineLen)
end;

procedure SetupYUVTables;
var
  i        : Longint;
  x,
  YCbBScale,
  YCrRScale,
  YCbGScale,
  YCrGScale: Double;
begin
(* general YUV (better: YCbCr) to RGB computation scheme:

   | R |   |  1.000000  1.402000  0.000000 |   | Y  |
   | G | = |  1.000000 -0.714136 -0.344136 | * | Cr |
   | B |   |  1.000000  0.000000  1.772000 |   | Cb |

  The following precomputed lookup tables are approximative due to rounding.
  Compared to the game-generated pictures, small differences (usually
  unnoticable) are possible.
 *)
	YCbBScale := 1.775 * 256;
	YCrRScale := 1.402 * 256;
	YCbGScale := 0.344 * 256;
	YCrGScale := 0.714 * 256;

  for i := 0 to 255 do
  begin
    x := i - 128;

		YCbB[i] := Trunc(YCbBScale * x);
		YCrR[i] := Trunc(YCrRScale * x);
		YCbG[i] := Trunc(YCbGScale * x);
		YCrG[i] := Trunc(YCrGScale * x);
		YTable[i] := i shl 8
  end
end;

function Clamp(i: Longint; Max: Longint): Longint;
begin
  if i < 0 then
    Result := 0
  else if i > Max then
    Result := Max
  else
    Result := i
end;

procedure YUVToRGB(YUVEntry: TYUVEntry; var RGBEntry: TRGBEntry);
var
  i : Byte;
  YY: Longint;
begin
  for i := 0 to 3 do
    with RGBEntry[i] do
      with YUVEntry do
      begin
        YY := YTable[Y[i]];
        R := Clamp(YY + YCrR[Cr], 65535) shr 8;
        G := Clamp(YY - YCbG[Cb] - YCrG[Cr], 65535) shr 8;
        B := Clamp(YY + YCbB[Cb], 65535) shr 8
      end
end;

procedure Convert(XMGName: string);
const
  cYUV  = 0;
  cVoid = 1;
  cRGB  = 2;
var
  i,
  RunLen   : Longint;
  Opcode,
  ColorMode: Byte;
  x, y     : Longword;
  YUVEntry : TYUVEntry;
  Entries  : array of TRGBEntry;
  BMP      : TBMPFile;
begin
  WriteLn(' (', XMGHeader.Width, ' x ', XMGHeader.Height, ')');

  XMGName := Prefix(XMGName) + '.bmp';

  OpenBMP(BMP, XMGName, XMGHeader.Width, XMGHeader.Height);

  y := 0;
  x := 0;
  RunLen := 0;
  ColorMode := cVoid;

  while not Eof(XMGFile) do
  begin
    BlockRead(XMGFile, OpCode, 1);

    (* determine entry type. Each opcode is interpreted as a bitfield of the
       format xxyyzzzz, where xx denotes the color mode (00, 01, 10) and yyzzzz
       denotes the number of 2x2 pixel blocks to be processed.
       In case xx equals 11, yy is the color mode and a second opcode aaaaaaaa
       has to be read. The number of blocks is then zzzzaaaaaaaa.
     *)
    case OpCode of
      $00..$3f: begin
                  ColorMode := cYUV;
                  RunLen := OpCode
                end;
      $40..$7f: begin
                  ColorMode := cVoid;
                  RunLen := OpCode and $3f
                end;
      $80..$bf: begin
                  ColorMode := cRGB;
                  RunLen := OpCode and $3f
                end;
      $c0..$fe: begin
                  // need to read further parameter
                  ColorMode := (OpCode shr 4) and $03;
                  RunLen := (OpCode and $0f) shl 8;

                  BlockRead(XMGFile, OpCode, 1);
                  RunLen := RunLen + OpCode
                end;
           $ff: Break
    end;

    // read entries
    SetLength(Entries, RunLen);
    case ColorMode of
      cVoid: for i := 0 to RunLen - 1 do
             begin
               Entries[i][0] := XMGHeader.VoidClr;
               Entries[i][1] := XMGHeader.VoidClr;
               Entries[i][2] := XMGHeader.VoidClr;
               Entries[i][3] := XMGHeader.VoidClr
             end;
      cYUV : for i := 0 to RunLen - 1 do
             begin
               BlockRead(XMGFile, YUVEntry, SizeOf(TYUVEntry));
               YUVToRGB(YUVEntry, Entries[i])
             end;
      cRGB : BlockRead(XMGFile, Entries[0], SizeOf(TRGBEntry) * RunLen)
    end;

    // draw entries
    for i := 0 to RunLen - 1 do
    begin
      with Entries[i][0] do
        WritePixelToBMP(BMP, x,     y    , R, G, B);
      with Entries[i][1] do
        WritePixelToBMP(BMP, x + 1, y    , R, G, B);
      with Entries[i][2] do
        WritePixelToBMP(BMP, x,     y + 1, R, G, B);
      with Entries[i][3] do
        WritePixelToBMP(BMP, x + 1, y + 1, R, G, B);

      Inc(x, 2);

      if x >= XMGHeader.Width then
      begin
        x := 0;
        Inc(y, 2)
      end
    end
  end;

  Finalize(Entries);
  CloseBMP(BMP)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  XMGName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    SetupYUVTables;

    repeat
      FileName := SearchRec.Name;
      XMGName := Dir + FileName;

      Write(XMGName);

      // if this file is accessible, analyze it
      if OpenXMGFile(XMGFile, XMGName) then
      begin
        if ReadHeader then
          Convert(XMGName)
        else
          WriteLn(' is not a valid (or supported) XMG file');
        Close(XMGFile)
      end
      else
        Writeln(XMGName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount >= 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

